
# (c) COPYRIGHT 1996 URI/MIT
# Please read the full copyright statement in the file COPYRIGH.  
#
# Authors:
#      jhrg,jimg       James Gallagher (jgallagher@gso.uri.edu)

# $Log: progress.tcl,v $
# Revision 1.2  1996/11/19 00:22:31  jimg
# Fixed the progress bar so that it no longer looks so silly.
#
# Revision 1.1  1996/11/18 23:14:59  jimg
# Added.
#

# Global variables used by the progress indictor.

# The previous position of the progress bar (used when PERCENT is -1; see the
# progress_bar procedure. This value forces the initial box to be displayed
# flush with the left-hand side of the widget.
set progress_bar_pos -1

# The id of the last rectangle displayed in the progress bar; used to erase.
set progress_bar_rect 0
set sign +

# Create the progress indicator window. Because the toplevel `.progress' is
# withdrawn from the window manager the progress indicator must be
# deiconified to be shown after calling this procedure.

proc progress_create {{text ""}} {
    global progress_geometry
    global font

    if {![winfo exists .progress]} {
	toplevel .progress
	wm withdraw .progress		;# Hide the new toplevel window

	wm geometry .progress $progress_geometry
	wm title .progress "dods transmission"

	frame .progress.f -width 60m -height 20m

	label .progress.f.icon -bitmap info 
	label .progress.f.msg -font $font -text $text -borderwidth 2 \
		-width 22 -anchor w

	canvas .progress.f.bar -relief sunken -width 56m -height 3m \
		-borderwidth 2

	pack .progress.f -padx 6 -pady 6 
	pack .progress.f.icon -side left -padx 20 -pady 20 -anchor n 
	pack .progress.f.msg -side top -padx 6 -pady 6 -fill y -expand true
	pack .progress.f.bar -side top -padx 6 -pady 6 -anchor s
    }
}

# Replace the current text in the progress indicator with NEW_TEXT.

proc progress_text {{new_text ""}} {
    .progress.f.msg configure -text $new_text
}

# Update progress bar to show percent transferred. If PERCENT is -1, then show
# a `bouncing rectangle' a la Netscape. If PERCENT is < -1 simply erase the
# bar and return.

proc progress_bar {{percent -2}} {
    global progress_bar_pos
    global progress_bar_rect
    global sign

    # Erase old bar.
    if {$progress_bar_rect != 0} {
	.progress.f.bar delete $progress_bar_rect
    }

    if {$percent < -1} {
	return
    } 

    # Use the pixel sizes (from the widget) rather than the millimeter sizes
    # from the globals.
    set bar_height [.progress.f.bar cget -height]
    set bar_width [.progress.f.bar cget -width]

    # Bouncing rectagle?
    if {$percent == -1} {

	# Calculate NEW_POS and possibly flip SIGN
	# NB: progress_bar_pos is the left side of the little rectabgle,
	# and NEW_POS is the right side. The Rectangle is BOX_WIDTH pixels
	# long.
	set box_width [expr $bar_width / 5]
	set incr_size [expr $box_width / 2]
	if {$progress_bar_pos == -1} {
	    set new_pos 0
	} else {
	    set new_pos [expr $progress_bar_pos $sign $incr_size]
	}
	
	if {[expr $new_pos + $box_width] > $bar_width} {
	    set sign -
	    set new_pos [expr $bar_width - $box_width]
	} elseif {$new_pos < 0} {
	    set sign +
	    set new_pos $box_width
	}
	
	# Add a fudge factor to make the bar fill the canvas widget. (Why?)
	#set new_pos [expr $new_pos $sign 4]
	set bar_height [expr $bar_height + 4]

	# Now draw the rectangle
	set progress_bar_rect [.progress.f.bar create \
		rectangle [expr $new_pos - 4] 0 \
		          [expr $new_pos + $box_width + 4] $bar_height \
		-fill Blue ]

	# Update PROGRESS_BAR_POS for next time.
	set progress_bar_pos $new_pos
    } else {
	set progress_bar_pos [expr $bar_width * ($percent / 100.0)]

	# Add fudge factor...
	set progress_bar_pos [expr $progress_bar_pos + 4]
	set bar_height [expr $bar_height + 4]

	set progress_bar_rect [.progress.f.bar create rectangle 0 0 \
		$progress_bar_pos $bar_height -fill Blue -width 0]
    }
}    

# Show the progress indicator window.

proc progress_popup {{text ""}} {
    if {![winfo exists .progress]} {
	progress_create $text
    }
    
    progress_text $text
    wm deiconify .progress
}

# Hide the progress indicator window.

proc progress_popdown {} {
    if {[winfo exists .progress]} {
	wm withdraw .progress
    }
}
   
# The interface to the progress indicator popup. Commands are fed to the
# popup using this procedure. Each takes the form `progress <command> ?arg?'
# where ?arg? is optional.

proc progress args {
    switch -regexp $args {
	"^popup*" {
	    progress_popup [lindex $args 1]
	    puts "OK"
	}
	"^popdown*$" {
	    global progress_bar_pos
	    set progress_bar_pos -1
	    progress_bar -2
	    progress_popdown
	    puts "OK"
	}
	"^text*" {
	    progress_text [lindex $args 1]
	    puts "OK"
	}
	"^bar*" {
	    progress_bar [lindex $args 1]
	    puts "OK"
	}
	default {
	    puts "Command unknown: $args"
	}
    }

    update
}
